unit Unit1;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Menus, CCR.Exif, StdCtrls, Jpeg;

type
  RGBSingle = record
    R, G, B: single;
  end;

  TForm1 = class(TForm)
    MainMenu1: TMainMenu;
    Datei1: TMenuItem;
    Open1: TMenuItem;
    Save1: TMenuItem;
    OpenDialog1: TOpenDialog;
    ProcessButton: TButton;
    SaveDialog1: TSaveDialog;
    procedure Open1Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure ProcessButtonClick(Sender: TObject);
    procedure Save1Click(Sender: TObject);
  private
    { Private-Deklarationen }
  public
    procedure resample(var Bitmap: TBitmap; factor: single);
  end;

var
  Form1: TForm1;
  JpegImageEx, JpegImageExNew: TJpegImageEx;
  NewThumb: TJpegImage;
  Bitmap: TBitmap;
  Filename: TFilename;

implementation

{$R *.dfm}

procedure TForm1.FormCreate(Sender: TObject);
begin
  JpegImageEx := TJpegImageEx.Create;
  JpegImageExNew := TJpegImageEx.Create;
  Bitmap := TBitmap.Create;
  NewThumb := TJpegImage.Create;
  NewThumb.PixelFormat := TJPEGPixelFormat(jf24Bit);
  NewThumb.CompressionQuality := 90;
end;

procedure TForm1.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  JpegImageEx.Free;
  JpegImageExNew.Free;
  Bitmap.Free;
  NewThumb.Free;
end;

procedure TForm1.Open1Click(Sender: TObject);
begin
  if OpenDialog1.Execute then begin
    Filename := OpenDialog1.FileName;
    JpegImageEx.LoadFromFile(Filename);
    ProcessButton.Visible := true;
  end;
end;

procedure TForm1.ProcessButtonClick(Sender: TObject);
begin
// do something with Bitmap
  Bitmap.Assign(JpegImageEx);
  Bitmap.Canvas.StretchDraw(Rect(0, 0, 1000, 1000), Bitmap);
  ProcessButton.Visible := false;
end;

procedure TForm1.Save1Click(Sender: TObject);
var
  factor: single;
  w, h: integer;
begin
  JpegImageExNew.Assign(JpegImageEx);
    // I need a new JpegImageEx in save-thread of my Application
  JpegImageExNew.Assign(Bitmap, [jaPreserveMetadata]);

  w := Bitmap.Width;
  h := Bitmap.Height;
  if w >= h then begin
            Factor := w/160;
          end else begin
            Factor := h/160;
          end;
  resample(Bitmap, Factor);
  // ...
  // here Bitmap is sharpend
  // ...
  // (resample and sharpen is better than to work with strechdraw)
  NewThumb.Assign(Bitmap);

  with JpegImageExNew.ExifData do begin
    Thumbnail := NewThumb;
    Orientation := toTopLeft;
    ThumbnailOrientation := toTopLeft;
    ExifImageWidth := w;
    ExifImageHeight := h;
  end;

  SaveDialog1.FileName := Filename;
  if SaveDialog1.Execute then JpegImageExNew.SaveToFile(SaveDialog1.FileName);
end;

procedure DownSample(var outBmp, inBmp: TBitmap);
const
	squareRootMax = 65280; // sqr(255.4995);
type
	squareRootArray = array[0..squareRootMax] of byte;
var
	factorX, faktorY: double;
	x_in_Weight, y_in_Weight: array of single;
	x_in_to_out, y_in_to_out: array of integer;
	x_in, y_in, x_out, y_out: integer;
	x1, y1: double;
	squareRoot: ^squareRootArray;
	bufLine, bufLine2, emptyLine: array of RGBSingle;
	areaScale, w, wx, wy, wy_scale, wy_inv_scale: single;
	inP, outP: PByteArray;
	sq_R, sq_G, sq_B: integer;
	ofs: integer;
	y_Carry: boolean;
begin
	factorX := outBmp.Width/inBmp.Width;
	faktorY := outBmp.Height/inBmp.Height;

	// Gewichte und Zuordnung vorausberechnen
	SetLength(x_in_Weight, inBmp.Width);
	SetLength(y_in_Weight, inBmp.Height);
	SetLength(x_in_to_out, inBmp.Width);
	SetLength(y_in_to_out, inBmp.Height);

	x_out := 0;
	for x_in := 0 to inBmp.Width - 1 do begin
		x_in_to_out[x_in] := x_out;
		x1 := factorX * (x_in + 1) - x_out;
		if x1 >= 1 then begin
			// bertrag ins nchste Pixel
			x_in_Weight[x_in] := 1 - (x1 - 1) / factorX;
			Inc(x_out, 1);
		end else
			x_in_Weight[x_in] := 1;
	end;

	y_out := 0;
	for y_in := 0 to inBmp.Height - 1 do begin
		y_in_to_out[y_in] := y_out;
		y1 := faktorY * (y_in + 1) - y_out;
		if y1 >= 1 then begin
			// bertrag ins nchste Pixel
			y_in_Weight[y_in] := 1 - (y1 - 1) / faktorY;
			Inc(y_out, 1);
		end else
			y_in_Weight[y_in] := 1;
	end;

	GetMem(squareRoot, squareRootMax + 1);
	try
		// Quadratwurzeln initialisieren zur schnelleren Berechnung
		for sq_R := 0 to squareRootMax do
			squareRoot[sq_R] := round(sqrt(sq_R));

		// resampling durchfhren
		SetLength(emptyLine, outBmp.Width + 1);
		for x_out := 0 to outBmp.Width do
			with emptyLine[x_out] do begin
				R := 0;
				G := 0;
				B := 0;
			end;
		bufLine := Copy(emptyLine);
		bufLine2 := Copy(emptyLine);
		areaScale := factorX * faktorY;

		for y_in := 0 to inBmp.Height - 1 do begin
			wy := y_in_Weight[y_in];
			wy_scale := wy * areaScale;
			wy_inv_scale := (1 - wy) * areaScale;
			y_Carry := (wy < 1);
			inP := inBmp.ScanLine[y_in];
			ofs := 0;
			for x_in := 0 to inBmp.Width - 1 do begin
				sq_R := inP[ofs + 2];
				sq_G := inP[ofs + 1];
				sq_B := inP[ofs];
				Inc(ofs, 3);
				sq_R := sq_R*sq_R;
				sq_G := sq_G*sq_G;
				sq_B := sq_B*sq_B;
				wx := x_in_Weight[x_in];
				w := wx * wy_scale;
				x_out := x_in_to_out[x_in];
				with bufLine[x_out] do begin
					R := R + sq_R * w;
					G := G + sq_G * w;
					B := B + sq_B * w;
				end;
				if (wx < 1) then begin
					// bertrag nach rechts
					w := (1 - wx) * wy_scale;
					with bufLine[x_out + 1] do begin
						R := R + sq_R * w;
						G := G + sq_G * w;
						B := B + sq_B * w;
					end;
					if y_Carry then begin
						// bertrag diagonal
						w := (1 - wx) * wy_inv_scale;
						with bufLine2[x_out + 1] do begin
							R := R + sq_R * w;
							G := G + sq_G * w;
							B := B + sq_B * w;
						end;
					end;
				end;
				if y_Carry then begin
					// bertrag nach unten
					w := wx * wy_inv_scale;
					with bufLine2[x_out] do begin
						R := R + sq_R * w;
						G := G + sq_G * w;
						B := B + sq_B * w;
					end;
				end;
			end;
			y_out := y_in_to_out[y_in];
			if (y_out < outBmp.Height) and
			   ((y_in = inBmp.Height - 1) or (y_in_to_out[y_in + 1] > y_out)) then begin
				// letzte Eingabezeile oder nchste Eingabezeile fllt in tiefere Ausgabezeile
				// gepufferte Zeile schreiben
				outP := outBmp.ScanLine[y_out];
				ofs := 0;
				for x_out := 0 to outBmp.Width - 1 do begin
					with bufLine[x_out] do begin
						outP[ofs + 2] := squareRoot[round(R)]; // round(sqrt(R));
						outP[ofs + 1] := squareRoot[round(G)]; // round(sqrt(G));
						outP[ofs] := squareRoot[round(B)]; // round(sqrt(B));
					end;
					Inc(ofs, 3);
				end;
				bufLine := bufLine2;
				bufLine2 := Copy(emptyLine);
			end;
		end;

	finally
		FreeMem(squareRoot);
	end;
end;

procedure UpSample(var outBmp, inBmp: TBitmap);
begin
// ...
end;

procedure TForm1.resample(var Bitmap: TBitmap; factor: single);
var
  help: TBitmap;
begin
  if factor = 1 then exit;
  try
    help := TBitmap.Create;
    help.Assign(Bitmap);
    Bitmap.SetSize(round(help.Width/factor), round(help.Height/factor));

    if factor>1 then
  		DownSample(Bitmap, help)
    else
  		UpSample(Bitmap, help);
  finally
    help.Free;
  end;
end;


end.
